@@ -38,6 +38,11 @@ module Brick.Main
   -- * Rendering cache management
   , invalidateCacheEntry
   , invalidateCache
+
+  -- * Renderer internals (for benchmarking)
+  , renderFinal
+  , getRenderState
+  , resetRenderState
   )
 where
 
@@ -180,8 +185,7 @@ runWithNewVty buildVty brickChan mUserChan app initialRS initialSt =
               Nothing -> readBChan brickChan
               Just uc -> readBrickEvent brickChan uc
             runInner rs st = do
-              (result, newRS) <- runVty vty readEvent app st (rs & observedNamesL .~ S.empty
-                                                                 & clickableNamesL .~ mempty)
+              (result, newRS) <- runVty vty readEvent app st (resetRenderState rs)
               case result of
                   SuspendAndResume act -> do
                       killThread pid
@@ -218,10 +222,11 @@ customMain buildVty mUserChan app initialAppState = do
                 InternalSuspendAndResume newRS action -> do
                     newAppState <- action
                     newVty <- buildVty
-                    run newVty newRS newAppState brickChan
+                    run newVty (newRS { renderCache = mempty }) newAppState brickChan
 
         emptyES = ES [] mempty
-        eventRO = EventRO M.empty initialVty mempty
+        emptyRS = RS M.empty mempty S.empty mempty mempty
+        eventRO = EventRO M.empty initialVty mempty emptyRS
     (st, eState) <- runStateT (runReaderT (runEventM (appStartEvent app initialAppState)) eventRO) emptyES
     let initialRS = RS M.empty (esScrollRequests eState) S.empty mempty []
     brickChan <- newBChan 20
@@ -303,7 +308,7 @@ runVty vty readEvent app appState rs = do
         _ -> return (e, firstRS, exts)
 
     let emptyES = ES [] mempty
-        eventRO = EventRO (viewportMap nextRS) vty nextExts
+        eventRO = EventRO (viewportMap nextRS) vty nextExts nextRS
 
     (next, eState) <- runStateT (runReaderT (runEventM (appHandleEvent app appState e'))
                                 eventRO) emptyES
@@ -373,6 +378,14 @@ withVty :: Vty -> (Vty -> IO a) -> IO a
 withVty vty useVty = do
     useVty vty `finally` shutdown vty
 
+getRenderState :: EventM n (RenderState n)
+getRenderState = EventM $ asks oldState
+
+resetRenderState :: RenderState n -> RenderState n
+resetRenderState s =
+    s & observedNamesL .~ S.empty
+      & clickableNamesL .~ mempty
+
 renderApp :: Vty -> App s e n -> s -> RenderState n -> IO (RenderState n, [Extent n])
 renderApp vty app appState rs = do
     sz <- displayBounds $ outputIface vty
@@ -479,7 +492,7 @@ halt = return . Halt
 
 -- | Suspend the event loop, save the terminal state, and run the
 -- specified action. When it returns an application state value, restore
--- the terminal state, redraw the application from the new state, and
--- resume the event loop.
+-- the terminal state, empty the rendering cache, redraw the application
+-- from the new state, and resume the event loop.
 suspendAndResume :: IO s -> EventM n (Next s)
 suspendAndResume = return . SuspendAndResume
