@@ -38,7 +38,9 @@ module Brick.Widgets.Core
 
   -- * Limits
   , hLimit
+  , hLimitPercent
   , vLimit
+  , vLimitPercent
   , setAvailableSize
 
   -- * Attribute management
@@ -239,12 +241,14 @@ takeColumns numCols (c:cs) =
             else ""
 
 -- | Make a widget from a string, but wrap the words in the input's
--- lines at the available width using the default wrapping settings.
+-- lines at the available width using the default wrapping settings. The
+-- input string should not contain escapes.
 strWrap :: String -> Widget n
 strWrap = strWrapWith defaultWrapSettings
 
 -- | Make a widget from a string, but wrap the words in the input's
 -- lines at the available width using the specified wrapping settings.
+-- The input string should not contain escapes.
 strWrapWith :: WrapSettings -> String -> Widget n
 strWrapWith settings t = txtWrapWith settings $ T.pack t
 
@@ -252,12 +256,14 @@ safeTextWidth :: T.Text -> Int
 safeTextWidth = V.safeWcswidth . T.unpack
 
 -- | Make a widget from text, but wrap the words in the input's lines at
--- the available width using the default wrapping settings.
+-- the available width using the default wrapping settings. The input
+-- text should not contain escapes.
 txtWrap :: T.Text -> Widget n
 txtWrap = txtWrapWith defaultWrapSettings
 
 -- | Make a widget from text, but wrap the words in the input's lines at
--- the available width using the specified wrapping settings.
+-- the available width using the specified wrapping settings. The input
+-- text should not contain escapes.
 txtWrapWith :: WrapSettings -> T.Text -> Widget n
 txtWrapWith settings s =
     Widget Greedy Fixed $ do
@@ -281,7 +287,8 @@ txtWrapWith settings s =
 -- The input string must not contain tab characters. If it does,
 -- interface corruption will result since the terminal will likely
 -- render it as taking up more than a single column. The caller should
--- replace tabs with the appropriate number of spaces as desired.
+-- replace tabs with the appropriate number of spaces as desired. The
+-- reinput string should not contain escapes.
 str :: String -> Widget n
 str s =
     Widget Fixed Fixed $ do
@@ -306,7 +313,8 @@ str s =
 -- The input string must not contain tab characters. If it does,
 -- interface corruption will result since the terminal will likely
 -- render it as taking up more than a single column. The caller should
--- replace tabs with the appropriate number of spaces as desired.
+-- replace tabs with the appropriate number of spaces as desired. The
+-- reinput text should not contain escapes.
 txt :: T.Text -> Widget n
 txt = str . T.unpack
 
@@ -414,16 +422,20 @@ fill ch =
 -- in the specified order (uppermost first). Defers growth policies to
 -- the growth policies of the contained widgets (if any are greedy, so
 -- is the box).
+{-# NOINLINE vBox #-}
 vBox :: [Widget n] -> Widget n
 vBox [] = emptyWidget
+vBox [a] = a
 vBox pairs = renderBox vBoxRenderer pairs
 
 -- | Horizontal box layout: put the specified widgets next to each other
 -- in the specified order (leftmost first). Defers growth policies to
 -- the growth policies of the contained widgets (if any are greedy, so
 -- is the box).
+{-# NOINLINE hBox #-}
 hBox :: [Widget n] -> Widget n
 hBox [] = emptyWidget
+hBox [a] = a
 hBox pairs = renderBox hBoxRenderer pairs
 
 -- | The process of rendering widgets in a box layout is exactly the
@@ -751,7 +763,22 @@ rewriteImage br (loRewrite, hiRewrite) old = rewriteHi . rewriteLo $ old where
 hLimit :: Int -> Widget n -> Widget n
 hLimit w p =
     Widget Fixed (vSize p) $
-      withReaderT (& availWidthL %~ (min w)) $ render $ cropToContext p
+      withReaderT (availWidthL %~ (min w)) $ render $ cropToContext p
+
+-- | Limit the space available to the specified widget to the specified
+-- percentage of available width, as a value between 0 and 100
+-- inclusive. Values outside the valid range will be clamped to the
+-- range endpoints. This is important for constraining the horizontal
+-- growth of otherwise-greedy widgets. This is non-greedy horizontally
+-- and defers to the limited widget vertically.
+hLimitPercent :: Int -> Widget n -> Widget n
+hLimitPercent w' p =
+    Widget Fixed (vSize p) $ do
+      let w = clamp 0 100 w'
+      ctx <- getContext
+      let usableWidth = ctx^.availWidthL
+          widgetWidth = round (toRational usableWidth * (toRational w / 100))
+      withReaderT (availWidthL %~ (min widgetWidth)) $ render $ cropToContext p
 
 -- | Limit the space available to the specified widget to the specified
 -- number of rows. This is important for constraining the vertical
@@ -760,7 +787,22 @@ hLimit w p =
 vLimit :: Int -> Widget n -> Widget n
 vLimit h p =
     Widget (hSize p) Fixed $
-      withReaderT (& availHeightL %~ (min h)) $ render $ cropToContext p
+      withReaderT (availHeightL %~ (min h)) $ render $ cropToContext p
+
+-- | Limit the space available to the specified widget to the specified
+-- percentage of available height, as a value between 0 and 100
+-- inclusive. Values outside the valid range will be clamped to the
+-- range endpoints. This is important for constraining the vertical
+-- growth of otherwise-greedy widgets. This is non-greedy vertically and
+-- defers to the limited widget horizontally.
+vLimitPercent :: Int -> Widget n -> Widget n
+vLimitPercent h' p =
+    Widget (vSize p) Fixed $ do
+      let h = clamp 0 100 h'
+      ctx <- getContext
+      let usableHeight = ctx^.availHeightL
+          widgetHeight = round (toRational usableHeight * (toRational h / 100))
+      withReaderT (availHeightL %~ (min widgetHeight)) $ render $ cropToContext p
 
 -- | Set the rendering context height and width for this widget. This
 -- is useful for relaxing the rendering size constraints on e.g. layer
@@ -1003,33 +1045,42 @@ viewport vpname typ p =
       reqs <- lift $ gets $ (^.rsScrollRequestsL)
       let relevantRequests = snd <$> filter (\(n, _) -> n == vpname) reqs
       when (not $ null relevantRequests) $ do
-          Just vp <- lift $ gets $ (^.viewportMapL.to (M.lookup vpname))
-          let updatedVp = applyRequests relevantRequests vp
-              applyRequests [] v = v
-              applyRequests (rq:rqs) v =
-                  case typ of
-                      Horizontal -> scrollTo typ rq (initialResult^.imageL) $ applyRequests rqs v
-                      Vertical -> scrollTo typ rq (initialResult^.imageL) $ applyRequests rqs v
-                      Both -> scrollTo Horizontal rq (initialResult^.imageL) $
-                              scrollTo Vertical rq (initialResult^.imageL) $
-                              applyRequests rqs v
-          lift $ modify (& viewportMapL %~ (M.insert vpname updatedVp))
-          return ()
+          mVp <- lift $ gets $ (^.viewportMapL.to (M.lookup vpname))
+          case mVp of
+              Nothing -> error $ "BUG: viewport: viewport name " <> show vpname <> " absent from viewport map"
+              Just vp -> do
+                  let updatedVp = applyRequests relevantRequests vp
+                      applyRequests [] v = v
+                      applyRequests (rq:rqs) v =
+                          case typ of
+                              Horizontal -> scrollTo typ rq (initialResult^.imageL) $ applyRequests rqs v
+                              Vertical -> scrollTo typ rq (initialResult^.imageL) $ applyRequests rqs v
+                              Both -> scrollTo Horizontal rq (initialResult^.imageL) $
+                                      scrollTo Vertical rq (initialResult^.imageL) $
+                                      applyRequests rqs v
+                  lift $ modify (& viewportMapL %~ (M.insert vpname updatedVp))
 
       -- If the sub-rendering requested visibility, update the scroll
       -- state accordingly
       when (not $ null $ initialResult^.visibilityRequestsL) $ do
-          Just vp <- lift $ gets $ (^.viewportMapL.to (M.lookup vpname))
-          let rqs = initialResult^.visibilityRequestsL
-              updateVp vp' rq = case typ of
-                  Both -> scrollToView Horizontal rq $ scrollToView Vertical rq vp'
-                  Horizontal -> scrollToView typ rq vp'
-                  Vertical -> scrollToView typ rq vp'
-          lift $ modify (& viewportMapL %~ (M.insert vpname $ foldl updateVp vp rqs))
+          mVp <- lift $ gets $ (^.viewportMapL.to (M.lookup vpname))
+          case mVp of
+              Nothing -> error $ "BUG: viewport: viewport name " <> show vpname <> " absent from viewport map"
+              Just vp -> do
+                  let rqs = initialResult^.visibilityRequestsL
+                      updateVp vp' rq = case typ of
+                          Both -> scrollToView Horizontal rq $ scrollToView Vertical rq vp'
+                          Horizontal -> scrollToView typ rq vp'
+                          Vertical -> scrollToView typ rq vp'
+                  lift $ modify (& viewportMapL %~ (M.insert vpname $ foldl updateVp vp rqs))
 
       -- If the size of the rendering changes enough to make the
       -- viewport offsets invalid, reset them
-      Just vp <- lift $ gets $ (^.viewportMapL.to (M.lookup vpname))
+      mVp <- lift $ gets $ (^.viewportMapL.to (M.lookup vpname))
+      vp <- case mVp of
+          Nothing -> error $ "BUG: viewport: viewport name " <> show vpname <> " absent from viewport map"
+          Just v -> return v
+
       let img = initialResult^.imageL
           fixTop v = if V.imageHeight img < v^.vpSize._2
                    then v & vpTop .~ 0
@@ -1044,7 +1095,10 @@ viewport vpname typ p =
       lift $ modify (& viewportMapL %~ (M.insert vpname (updateVp vp)))
 
       -- Get the viewport state now that it has been updated.
-      Just vpFinal <- lift $ gets (M.lookup vpname . (^.viewportMapL))
+      mVpFinal <- lift $ gets (M.lookup vpname . (^.viewportMapL))
+      vpFinal <- case mVpFinal of
+          Nothing -> error $ "BUG: viewport: viewport name " <> show vpname <> " absent from viewport map"
+          Just v -> return v
 
       -- Then perform a translation of the sub-rendering to fit into the
       -- viewport
@@ -1177,6 +1231,7 @@ visibleRegion vrloc sz p =
 -- | Horizontal box layout: put the specified widgets next to each other
 -- in the specified order. Defers growth policies to the growth policies
 -- of both widgets.  This operator is a binary version of 'hBox'.
+{-# NOINLINE (<+>) #-}
 (<+>) :: Widget n
       -- ^ Left
       -> Widget n
@@ -1187,9 +1242,21 @@ visibleRegion vrloc sz p =
 -- | Vertical box layout: put the specified widgets one above the other
 -- in the specified order. Defers growth policies to the growth policies
 -- of both widgets.  This operator is a binary version of 'vBox'.
+{-# NOINLINE (<=>) #-}
 (<=>) :: Widget n
       -- ^ Top
       -> Widget n
       -- ^ Bottom
       -> Widget n
 (<=>) a b = vBox [a, b]
+
+{-# RULES
+"baseHbox" forall a b   . a <+> b                 = hBox [a, b]
+"hBox2"    forall as bs . hBox [hBox as, hBox bs] = hBox (as ++ bs)
+"hboxL"    forall as b  . hBox [hBox as, b]       = hBox (as ++ [b])
+"hboxR"    forall a bs  . hBox [a, hBox bs]       = hBox (a : bs)
+"baseVbox" forall a b   . a <=> b                 = vBox [a, b]
+"vBox2"    forall as bs . vBox [vBox as, vBox bs] = vBox (as ++ bs)
+"vboxL"    forall as b  . vBox [vBox as, b]       = vBox (as ++ [b])
+"vboxR"    forall a bs  . vBox [a, vBox bs]       = vBox (a : bs)
+#-}
